/*************************************************
* Unix EntropySource Source File                 *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#ifndef _XOPEN_SOURCE
  #define _XOPEN_SOURCE 500
#endif
#ifndef _XOPEN_SOURCE_EXTENDED
  #define _XOPEN_SOURCE_EXTENDED 1
#endif

#include <botan/es_unix.h>
#include <botan/util.h>
#include <stdio.h>

namespace Botan {

namespace {

bool Unix_Program_Cmp(const Unix_Program& a, const Unix_Program& b)
   { return (a.priority < b.priority); }

}

/*************************************************
* Unix_EntropySource Constructor                 *
*************************************************/
Unix_EntropySource::Unix_EntropySource(bool add_defaults)
   {
   if(add_defaults)
      add_default_sources();
   }

/*************************************************
* Unix_EntropySource Constructor                 *
*************************************************/
Unix_EntropySource::Unix_EntropySource(const Unix_Program srcs[], u32bit count,
                                       bool add_defaults)
   {
   add_sources(srcs, count);
   if(add_defaults)
      add_default_sources();
   }

/*************************************************
* Add the default sources                        *
*************************************************/
void Unix_EntropySource::add_default_sources()
   {
   u32bit count = 0;
   while(true)
      {
      if(DEFAULT_SOURCES[count].name == "") break;
      count++;
      }
   add_sources(DEFAULT_SOURCES, count);
   }

/*************************************************
* Fast Poll                                      *
*************************************************/
void Unix_EntropySource::do_fast_poll()
   {
   static const u32bit MAX_PRIORITY = 2;
   gather(MAX_PRIORITY, 512);
   }

/*************************************************
* Slow Poll                                      *
*************************************************/
void Unix_EntropySource::do_slow_poll()
   {
   static const u32bit MAX_PRIORITY = 4;
   gather(MAX_PRIORITY, 2048);
   }

/*************************************************
* Gather Entropy From Several Unix_Programs      *
*************************************************/
void Unix_EntropySource::gather(u32bit prio, u32bit length)
   {
   u32bit got = 0;
   for(u32bit j = 0; j != sources.size(); j++)
      {
      if((sources[j].priority > prio) && (got > length))
         break;
      if(got > 4*length)
         break;
      if(sources[j].working)
         got += gather_entropy(sources[j]);
      }
   }

/*************************************************
* Gather Entropy from a Unix_Program             *
*************************************************/
u32bit Unix_EntropySource::gather_entropy(Unix_Program& src)
   {
   static const u32bit MIN_OUTPUT = 10;
   static const u32bit MAX_READ = 8*1024;

   std::string command = src.name + " " + src.args + " 2>/dev/null";

   SecureVector<byte> output(DEFAULT_BUFFERSIZE);
   u32bit total_output = 0;

   timestamp();

   FILE* pipe = popen(command.c_str(), "r");
   if(!pipe)
      throw Stream_IO_Error("Unix_EntropySource: popen failed");
   while(!ferror(pipe) && !feof(pipe) && total_output <= MAX_READ)
      {
      u32bit got = fread(output, 1, output.size(), pipe);
      total_output += got;
      add_bytes(output, output.size());
      }

   if(pclose(pipe) == 127*256)   src.working = false;
   if(total_output < MIN_OUTPUT) src.working = false;

   timestamp();

   return total_output;
   }

/*************************************************
* Add a timestamp value to the buffer            *
*************************************************/
void Unix_EntropySource::timestamp()
   {
   u64bit time = system_clock();
   add_bytes(&time, sizeof(u64bit));
   }

/*************************************************
* Add sources to the list                        *
*************************************************/
void Unix_EntropySource::add_sources(const Unix_Program srcs[], u32bit count)
   {
   sources.insert(sources.end(), srcs, srcs + count);
   std::sort(sources.begin(), sources.end(), Unix_Program_Cmp);
   }

}
