/*************************************************
* Initalization Function Source File             *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/init.h>
#include <botan/allocate.h>
#include <botan/rng.h>
#include <botan/mutex.h>
#include <botan/timers.h>
#include <botan/look_add.h>
#include <botan/randpool.h>
#include <botan/x917_rng.h>
#include <botan/util.h>
#include <map>

#if defined(BOTAN_EXT_MUTEX_PTHREAD)
  #include <botan/mux_pthr.h>
#endif

#if defined(BOTAN_EXT_ALLOC_MMAP)
  #include <botan/mmap_mem.h>
#endif

#if defined(BOTAN_EXT_ALLOC_MLOCK)
  #include <botan/lock_mem.h>
#endif

#if defined(BOTAN_EXT_TIMER_HARDWARE)
  #include <botan/tm_hard.h>
#elif defined(BOTAN_EXT_TIMER_POSIX)
  #include <botan/tm_posix.h>
#elif defined(BOTAN_EXT_TIMER_UNIX)
  #include <botan/tm_unix.h>
#endif

namespace Botan {

/*************************************************
* Library Initalization                          *
*************************************************/
LibraryInitializer::LibraryInitializer(const std::string& arg_string)
   {
   Init::initialize(arg_string);
   }

/*************************************************
* Library Shutdown                               *
*************************************************/
LibraryInitializer::~LibraryInitializer()
   {
   Init::deinitialize();
   }

namespace {

/*************************************************
* Check if an option is set in the argument      *
*************************************************/
bool arg_set(const std::vector<std::string>& args,
             const std::string& option)
   {
   if(std::find(args.begin(), args.end(), option) != args.end())
      return true;
   return false;
   }

/*************************************************
* Register a mutex type, if possible             *
*************************************************/
bool set_mutex(bool must_have_mutex)
   {
   if(!must_have_mutex) { set_mutex_type(0); return true; }

#if defined(BOTAN_EXT_MUTEX_PTHREAD)
   set_mutex_type(new Pthread_Mutex);
   return true;
#else
   return false;
#endif
   }

/*************************************************
* Register a high resolution timer, if possible  *
*************************************************/
void set_timer()
   {
#if defined(BOTAN_EXT_TIMER_HARDWARE)
   set_timer_type(new Hardware_Timer);
#elif defined(BOTAN_EXT_TIMER_POSIX)
   set_timer_type(new POSIX_Timer);
#elif defined(BOTAN_EXT_TIMER_UNIX)
   set_timer_type(new Unix_Timer);
#endif
   }

/*************************************************
* Register a more secure allocator, if possible  *
*************************************************/
void set_safe_allocator()
   {
#if defined(BOTAN_EXT_ALLOC_MLOCK)
   add_allocator_type("mlock", new MemoryLocking_Allocator);
   set_default_allocator("mlock");
#endif

#if defined(BOTAN_EXT_ALLOC_MMAP)
   add_allocator_type("mmap", new MemoryMapping_Allocator);
   set_default_allocator("mmap");
#endif
   }

}

/*************************************************
* Internal startup/shutdown functions            *
*************************************************/
extern void initalize_memory_subsystem();
extern void shutdown_memory_subsystem();
extern void create_global_bigints();
extern void delete_global_bigints();
extern void init_lookup_tables();
extern void destroy_lookup_tables();
extern void load_algorithms_list();
extern void set_default_options();

namespace Init {

/*************************************************
* Library Initalization                          *
*************************************************/
void initialize(const std::string& arg_string)
   {
   std::vector<std::string> args = split_on(arg_string, ' ');

   if(!set_mutex(arg_set(args, "thread_safe")))
      throw Exception("LibraryInitializer: thread safety impossible");

   initalize_memory_subsystem();
   create_global_bigints();
   init_lookup_tables();
   set_default_options();

   if(arg_set(args, "secure_memory"))
      set_safe_allocator();
   set_timer();

   load_algorithms_list();
   setup_global_rng(new Randpool, new ANSI_X917_RNG);

   Global_RNG::seed();
   }

/*************************************************
* Library Shutdown                               *
*************************************************/
void deinitialize()
   {
   destroy_global_rng();
   destroy_lookup_tables();
   delete_global_bigints();
   set_timer_type(0);
   shutdown_memory_subsystem();
   }

}

}
