/*************************************************
* Main MP Algorithms Source File                 *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/mp_core.h>
#include <string.h>

extern "C" {

/*************************************************
* Two Operand Addition Without Carry             *
*************************************************/
word bigint_add2_nc(word* const x, u32bit x_size,
                    const word* y, u32bit y_size)
   {
   word carry = 0;

   for(u32bit j = 0; j != y_size; j++)
      {
      dword sum = ((dword)x[j] + y[j]) + carry;
      x[j] = MP_LOW_WORD(sum);
      carry = MP_HIGH_WORD(sum);
      }

   if(!carry)
      return 0;

   for(u32bit j = y_size; j != x_size; j++)
      {
      x[j]++;
      if(x[j])
         return 0;
      }
   return 1;
   }

/*************************************************
* Two Operand Addition                           *
*************************************************/
void bigint_add2(word* const x, u32bit x_size,
                 const word* y, u32bit y_size)
   {
   x[x_size] += bigint_add2_nc(x, x_size, y, y_size);
   }

/*************************************************
* Three Operand Addition                         *
*************************************************/
void bigint_add3(word* const z,
                 const word* x, u32bit x_size,
                 const word* y, u32bit y_size)
   {
   if(x_size < y_size) { bigint_add3(z, y, y_size, x, x_size); return; }

   word carry = 0;
   for(u32bit j = 0; j != y_size; j++)
      {
      dword sum = ((dword)x[j] + y[j]) + carry;
      z[j] = MP_LOW_WORD(sum);
      carry = MP_HIGH_WORD(sum);
      }

   for(u32bit j = y_size; j != x_size; j++)
      {
      if(carry == 0)
         {
         memcpy(z + j, x + j, (x_size - j) * sizeof(word));
         break;
         }
      z[j] = x[j] + carry;
      carry = ((x[j] == MP_WORD_MAX) ? 1 : 0);
      }
   z[x_size] = carry;
   }

/*************************************************
* Two Operand Subtraction                        *
*************************************************/
void bigint_sub2(word* const x, u32bit x_size,
                 const word* y, u32bit y_size)
   {
   const dword MP_RADIX = (dword)1 << MP_WORD_BITS;

   word borrow = 0;
   for(u32bit j = 0; j != y_size; j++)
      {
      dword temp = ((MP_RADIX + x[j]) - y[j]) - borrow;
      borrow = ((temp < MP_RADIX) ? 1 : 0);
      x[j] = (word)(temp - MP_RADIX);
      }
   for(u32bit j = y_size; j != x_size; j++)
      {
      if(borrow == 0)
         break;
      dword temp = (MP_RADIX + x[j]) - borrow;
      borrow = ((temp < MP_RADIX) ? 1 : 0);
      x[j] = (word)(temp - MP_RADIX);
      }
   }

/*************************************************
* Three Operand Subtraction                      *
*************************************************/
void bigint_sub3(word* const z,
                 const word* x, u32bit x_size,
                 const word* y, u32bit y_size)
   {
   const dword MP_RADIX = (dword)1 << MP_WORD_BITS;

   word borrow = 0;
   for(u32bit j = 0; j != y_size; j++)
      {
      dword temp = ((MP_RADIX + x[j]) - y[j]) - borrow;
      borrow = ((temp < MP_RADIX) ? 1 : 0);
      z[j] = (word)(temp - MP_RADIX);
      }
   for(u32bit j = y_size; j != x_size; j++)
      {
      if(borrow == 0)
         {
         memcpy(z + j, x + j, (x_size - j) * sizeof(word));
         break;
         }
      dword temp = (MP_RADIX + x[j]) - borrow;
      borrow = ((temp < MP_RADIX) ? 1 : 0);
      z[j] = (word)(temp - MP_RADIX);
      }
   }

/*************************************************
* Two Operand Linear Multiply                    *
*************************************************/
void bigint_linmul2(word* const x, u32bit x_size, word y)
   {
   word carry = 0;
   for(u32bit j = 0; j != x_size; j++)
      {
      dword product = (dword)x[j] * y + carry;
      x[j] = MP_LOW_WORD(product);
      carry = MP_HIGH_WORD(product);
      }
   x[x_size] = carry;
   }

/*************************************************
* Three Operand Linear Multiply                  *
*************************************************/
void bigint_linmul3(word* const z,
                    const word* x, u32bit x_size, word y)
   {
   word carry = 0;
   for(u32bit j = 0; j != x_size; j++)
      {
      dword product = (dword)x[j] * y + carry;
      z[j] = MP_LOW_WORD(product);
      carry = MP_HIGH_WORD(product);
      }
   z[x_size] = carry;
   }

}
