/*************************************************
* X.509 Public Key Source File                   *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/x509_key.h>
#include <botan/filters.h>
#include <botan/asn1_obj.h>
#include <botan/pk_algs.h>
#include <botan/oids.h>
#include <botan/pem.h>
#include <memory>

namespace Botan {

/*************************************************
* Return the default NULL algorithm parameters   *
*************************************************/
SecureVector<byte> X509_PublicKey::DER_encode_params() const
   {
   DER_Encoder encoder;
   DER::encode_null(encoder);
   return encoder.get_contents();
   }

/*************************************************
* Check to make sure the parameters are NULL     *
*************************************************/
void X509_PublicKey::BER_decode_params(DataSource& source)
   {
   BER_Decoder decoder(source);
   if(decoder.more_items())
      BER::decode_null(decoder);
   }

namespace X509 {

namespace {

/*************************************************
* Extract the fields of a subjectPublicKeyInfo   *
*************************************************/
void X509_extract_info(DataSource& source, AlgorithmIdentifier& alg_id,
                       SecureVector<byte>& x509_key)
   {
   BER_Decoder decoder(source);
   BER_Decoder sequence = BER::get_subsequence(decoder);
   BER::decode(sequence, alg_id);
   BER::decode(sequence, x509_key, BIT_STRING);
   sequence.verify_end();
   }

/*************************************************
* Decode a BER encoded X.509 public key          *
*************************************************/
void X509_BER_decode(X509_PublicKey& key, DataSource& source)
   {
   try {
      AlgorithmIdentifier alg_id;
      SecureVector<byte> x509_key;

      X509_extract_info(source, alg_id, x509_key);

      if(!key.valid_oid(alg_id.oid))
         throw Decoding_Error("X.509: OID mismatch, got " +
                              alg_id.oid.as_string() + ", expected " +
                              key.get_oid().as_string());

      Pipe output;
      output.process_msg(alg_id.parameters);
      output.process_msg(x509_key);
      key.BER_decode_params(output);
      output.set_default_msg(1);
      key.BER_decode_pub(output);
      }
   catch(Decoding_Error)
      {
      throw Decoding_Error("X.509 public key decoding failed");
      }
   }

}

/*************************************************
* DER or PEM encode a X.509 public key           *
*************************************************/
void encode(const X509_PublicKey& key, Pipe& pipe, X509_Encoding encoding)
   {
   DER_Encoder encoder;
   AlgorithmIdentifier alg_id(key.get_oid(), key.DER_encode_params());

   encoder.start_sequence();
   DER::encode(encoder, alg_id);
   DER::encode(encoder, key.DER_encode_pub(), BIT_STRING);
   encoder.end_sequence();

   SecureVector<byte> der = encoder.get_contents();
   if(encoding == PEM)
      pipe.write(pem_encode(der, der.size(), "PUBLIC KEY"));
   else
      pipe.write(der);
   }

/*************************************************
* PEM encode a X.509 public key                  *
*************************************************/
std::string PEM_encode(const X509_PublicKey& key)
   {
   Pipe pem;
   pem.start_msg();
   encode(key, pem, PEM);
   pem.end_msg();
   return pem.read_all_as_string();
   }

/*************************************************
* Decode a BER or PEM encoded X.509 public key   *
*************************************************/
void decode(X509_PublicKey& key, DataSource& source, X509_Encoding encoding)
   {
   if(encoding == PEM)
      {
      Pipe ber;
      ber.process_msg(pem_decode_check_label(source, "PUBLIC KEY"));
      X509_BER_decode(key, ber);
      }
   else
      X509_BER_decode(key, source);
   }

/*************************************************
* Extract a public key and return it             *
*************************************************/
X509_PublicKey* load_key(DataSource& source, X509_Encoding encoding)
   {
   try {
      AlgorithmIdentifier alg_id;
      SecureVector<byte> x509_key;

      if(encoding == PEM)
         {
         Pipe ber;
         ber.process_msg(pem_decode_check_label(source, "PUBLIC KEY"));
         X509_extract_info(ber, alg_id, x509_key);
         }
      else
         X509_extract_info(source, alg_id, x509_key);

      const std::string alg_name = OIDS::lookup(alg_id.oid);
      if(alg_name == "")
         throw Decoding_Error("Unknown algorithm OID: " +
                              alg_id.oid.as_string());

      std::auto_ptr<X509_PublicKey> key(get_public_key(alg_name));
      if(!key.get())
         throw Decoding_Error("Unknown PK algorithm/OID: " + alg_name + ", " +
                              alg_id.oid.as_string());

      Pipe output;
      output.process_msg(alg_id.parameters);
      output.process_msg(x509_key);
      key->BER_decode_params(output);
      output.set_default_msg(1);
      key->BER_decode_pub(output);

      return key.release();
      }
   catch(Decoding_Error)
      {
      throw Decoding_Error("X.509 public key decoding failed");
      }
   }

/*************************************************
* Extract a public key and return it             *
*************************************************/
X509_PublicKey* load_key(const SecureVector<byte>& pub_key)
   {
   DataSource_Memory source(pub_key);
   return X509::load_key(source, RAW_BER);
   }

/*************************************************
* Find the allowable key constraints             *
*************************************************/
Key_Constraints find_constraints(const X509_PublicKey& pub_key,
                                 Key_Constraints limits)
   {
   const X509_PublicKey* key = &pub_key;
   u32bit constraints = 0;

   if(dynamic_cast<const PK_Encrypting_Key*>(key))
      constraints |= KEY_ENCIPHERMENT;

   if(dynamic_cast<const PK_Key_Agreement_Key*>(key))
      constraints |= KEY_AGREEMENT;

   if(dynamic_cast<const PK_Verifying_wo_MR_Key*>(key) ||
      dynamic_cast<const PK_Verifying_with_MR_Key*>(key))
      constraints |= DIGITAL_SIGNATURE | NON_REPUDATION;

   if(limits)
      constraints &= limits;

   return Key_Constraints(constraints);
   }

}

}
